var sys = {
    allBookmarks: null,
    getAllBookmarks: function (func) {
        chrome.bookmarks.getTree(function (tree) {
            sys.allBookmarks = JSON.parse(JSON.stringify(tree[0].children));
            func && func(tree[0].children);
        });
    },
    getChildren: function (nodeID, nodes) {
        var res = [],
            nodes = (nodes === undefined) ? sys.allBookmarks : nodes;
        for (var i = 0, node; node = nodes[i++];) {
            if (Array.isArray(node.children)) {
                if (node.id == nodeID) {
                    res = node.children;
                } else {
                    res = sys.getChildren(nodeID, node.children);
                }
            }
            if (Array.isArray(res) && res.length > 0) {
                return res;
            }
        }
        return res;
    },
    search: function (query, func) {
        chrome.bookmarks.search(query, function (res) {
            func && func(res);
        });
    },
    getStorage: function (key) {
        return localStorage[key] ? localStorage[key] : undefined;
    },
    setStorage: function (key, value) {
        localStorage[key] = value;
    }
};

var ui = {
    $: function (id) {
        return document.getElementById(id);
    },
    create: function (tag) {
        return document.createElement(tag);
    },
    remove: function (element) {
        return element && element.parentNode && element.parentNode.removeChild(element);
    },
    listen: function (element, funcName, func) {
        element.addEventListener ? element.addEventListener(funcName, func, false) : element.attachEvent(
            "on" + funcName, func);
    },
    unlisten: function (element, funcName, func) {
        element.removeEventListener ? element.removeEventListener(funcName, func, false) : element
            .detachEvent("on" + funcName, func);
    },
    inArray: function (needle, array) {
        var len = array.length;
        for (var i = 0; i < len; i++) {
            if (array[i] === needle) {
                return true;
            }
        }
        return false;
    },
    getMousePos: function () {
        var evt = arguments[0] || window.event;
        return {
            x: evt.clientX,
            y: evt.clientY
        };
    },
    getItemID: function (item) {
        return item.id.split('-').pop().toString();
    }
};

var actions = {
    openItem: function (item, isSelected) {
        chrome.tabs.create({
            "url": item.url,
            "selected": isSelected
        });
    },
    openItemsInWindow: function (items, isIncognito) {
        chrome.extension.sendRequest({
            cmd: 'openItemsInWindow',
            data: items,
            isIncognito: isIncognito
        });
    },
    removeItem: function (item) {
        chrome.bookmarks.remove(ui.getItemID(item), function () {
            if (searchValue) {
                buildSearchResult(searchValue);
            } else {
                initTree();
            }
        });
    },
    removeTree: function (item) {
        var id = ui.getItemID(item);
        updateOpenFolderIDs(id);
        chrome.bookmarks.removeTree(id, function () {
            initTree();
        });
    },
    renameItem: function () {

    },
    manager: function () {
        chrome.tabs.create({
            // "url": 'chrome-extension://eemcgdkfndhakfknompkggombfjjjeno/main.html'
            "url": 'chrome://bookmarks/'
        });
    }
};

var floor = 0,
    index = 0,
    defaultFindID = "1", //Ĭ�ϴ򿪡��ղؼС�
    allShowItemsCount = 0,
    maxItemsMargin = 8,
    maxItemsCount = 25,
    unitItemHeight = 22.6,
    minWindowHeight = 200,
    searchValue = '',
    findIDStr = '',
    openFolderIDs = [],
    items,
    keyNames = {
        0: 'left',
        1: 'middle',
        2: 'right'
    },
    resizer = ui.$('resize'),
    tree = ui.$('bookmarks'),
    searchInput = ui.$('searchInput'),
    resizeHandle = false,
    bodyWidth,
    screenX;

function createItem(item, type, status, op) {
    allShowItemsCount++;
    var itemImg = status ? ('img/' + type + '-' + status + '.png') : ('chrome://favicon/' + item.url),
        li = ui.create('li');
    li.id = item.parentId + '-' + item.id;
    li.parentId = item.parentId;
    li.url = item.url;
    li.type = type;
    li.status = status;
    li.floor = floor;
    li.index = ++index;
    li.style.paddingLeft = 20 * floor + "px";
    //li.title = item.parentId+ "-" +item.id + '-' + index + '-' + type + '-' + status;
    li.title = item.url ? item.url : '';
    li.innerHTML = " <img src=" + itemImg + " /> " + item.title;
    li.onclick = function (e) {
        var self = this,
            itemID = ui.getItemID(self),
            keyName = keyNames[e.button],
            isSelectedOpenItem = (keyName == 'left') ? true : false;
        if (self.type == 'folder') {
            //�м���
            if (keyName == 'middle') {
                sys.getChildren(itemID).forEach(function (child) {
                    if (!Array.isArray(child.children)) {
                        actions.openItem(child, isSelectedOpenItem);
                    }
                });
            }
            //�����
            if (self.status == 'close') { //��  �۵����ļ���
                //�����ȹر�֮ǰ�򿪵��ļ��У����򽫵��´������
                //���ҵ�Ҫ�رյ��ļ���ID
                var i = 0,
                    curID = null,
                    pid = self.parentId;
                while (pid && (curID = openFolderIDs[i++])) {
                    if (curID == pid && openFolderIDs[i]) {
                        //�ر������ǵ�ǰĿ¼���Ѵ��ļ���
                        closeFolder(pid + '-' + openFolderIDs[i]);
                        break;
                    }
                }

                //�����´��ļ���ID
                openFolderIDs.push(itemID);

                //�����ļ����ڵĽڵ�
                index = self.index;
                floor = self.floor + 1;
                var count = 0,
                    children = sys.getChildren(itemID),
                    childOP = self.nextSibling ? {
                        pos: 'front',
                        nextSiblingId: self.nextSibling.id
                    } : {
                        pos: 'behind'
                    };
                while (child = children[count++]) {
                    recursiveCreateItem(child, childOP);
                }

                //�ı��ļ���״̬
                self.innerHTML = self.innerHTML.replace(/folder-close/, "folder-open");
                self.status = 'open';
                sys.setStorage('lastOpenFolderID', itemID);

            } else { //�ر�  չ�����ļ���
                closeFolder(self.id);
            }

        } else { //���Ӳ���
            actions.openItem(self, isSelectedOpenItem);
        }
    }

    if (op.pos == 'front') {
        tree.insertBefore(li, ui.$(op.nextSiblingId));
    } else {
        tree.appendChild(li);
    }
}

function createSearchItem(item) {
    var itemImg = 'chrome://favicon/' + item.url,
        li = ui.create('li');
    li.id = item.parentId + '-' + item.id;
    li.url = item.url;
    li.type = 'file';
    //li.title = item.parentId+ "-" +item.id ;
    li.title = item.url ? item.url : '';
    li.innerHTML = " <img src=" + itemImg + " /> " + item.title;
    li.onclick = function (e) {
        var self = this,
            keyName = keyNames[e.button],
            isSelectedOpenItem = (keyName == 'left') ? true : false;
        actions.openItem(self, isSelectedOpenItem);
    }
    tree.appendChild(li);
}

function buildSearchResult(query) {
    sys.search(query, function (items) {
        if (query) {
            menu.hideMenu();
            tree.innerHTML = '';
            resizer.style.height = document.body.style.height = maxItemsCount * unitItemHeight + 'px';
            items.forEach(function (item) {
                createSearchItem(item);
            });
        } else {
            initTree();
        }
    });
}

function closeFolder(id) {
    var self = ui.$(id),
        selfID = ui.getItemID(self);
    recursiveDeleteItem(selfID);
    self.innerHTML = self.innerHTML.replace(/folder-open/, "folder-close");
    self.status = 'close';
    updateOpenFolderIDs(selfID);
}

//�Ӵ򿪵��ļ����б��У� ȥ����ǰ�رյ��ļ��м����ļ���ID
function updateOpenFolderIDs(closeID) {
    var i = 0,
        curOpenID = null;
    while (curOpenID = openFolderIDs[i]) {
        if (curOpenID == closeID) {
            openFolderIDs.splice(i);
            break;
        }
        i++; //ע��i++��λ��
    }
    sys.setStorage('lastOpenFolderID', openFolderIDs.slice(-1));
}

function recursiveDeleteItem(parentId) {
    for (var i = 0, item; item = items[i++];) {
        var thisParentID = item.id ? item.parentId : 'undefined';
        if (parentId == thisParentID) {
            if (item.type == 'folder') {
                var curParentID = item.id ? ui.getItemID(item) : '';
                recursiveDeleteItem(curParentID);
            }
            ui.remove(item);
            i--; //getElementsByTag list index will modify when removed
            allShowItemsCount--;
        }
    }
}

function recursiveCreateItem(obj, op) {
    if (Array.isArray(obj.children)) {
        var isOpenFolder = ui.inArray(obj.id, openFolderIDs),
            folderStatus = isOpenFolder ? 'open' : 'close';
        createItem(obj, 'folder', folderStatus, op);
        if (isOpenFolder) {
            floor++;
            obj.children && obj.children.forEach(function (item) {
                recursiveCreateItem(item, op);
            });
            floor--;
        }
    } else {
        createItem(obj, 'file', '', op);
    }
}

function recursiveFindOpenFolder(obj, findID) {
    if (obj.id == findID) {
        return obj.id;
    } else if (Array.isArray(obj.children)) {
        var res = false;
        for (var i in obj.children) {
            if (Array.isArray(obj.children[i].children) && (res = recursiveFindOpenFolder(obj.children[i], findID))) {
                return obj.id + ',' + res;
            }
        }
    }
}

function findOpenFolders(dataSet) {
    var findIDStr = "0",
        findRes = false,
        lastOpenFolderID = sys.getStorage('lastOpenFolderID'),
        findID = lastOpenFolderID ? lastOpenFolderID : defaultFindID;

    for (var idx in dataSet) {
        var item = dataSet[idx];
        if (Array.isArray(item.children)) {
            if (findRes = recursiveFindOpenFolder(item, findID)) {
                findIDStr += ',' + findRes;
                break;
            }
        }
    }

    //���û���ҵ����򿪵��ļ��У����Ĭ���ļ���
    if (findIDStr == "0") {
        findIDStr += ',' + defaultFindID;
    }
    return findIDStr.split(',');
}

function initTree() {

    //��ʼ����������
    floor = 0;
    index = 0;
    allShowItemsCount = 0;
    searchValue = '';
    openFolderIDs = [];
    items = null;
    resizeHandle = false;

    sys.getAllBookmarks(function (bookmarks) {
        var maxHeight = 0,
            bodyWidth;
        tree.innerHTML = "";
        openFolderIDs = findOpenFolders(bookmarks);
        bookmarks.forEach(function (item) {
            recursiveCreateItem(item, {
                pos: 'behind',
                index: ''
            });
        });
        items = tree.getElementsByTagName('li');
        if (allShowItemsCount > maxItemsCount) {
            maxHeight = maxItemsCount * unitItemHeight;
        } else {
            maxHeight = allShowItemsCount * unitItemHeight + maxItemsMargin;
        }
        resizer.style.height = document.body.style.height = Math.max(minWindowHeight, maxHeight) + 'px';
        toolbar.resetTop(maxHeight - 30);
        bodyWidth = sys.getStorage('bodyWidth');
        if (bodyWidth) {
            document.body.style.width = bodyWidth + 'px';
            toolbar.resetWidth(bodyWidth);
        }
        resizer.style.left = (bodyWidth ? bodyWidth : document.body.getBoundingClientRect().width) - 5 + 'px';
        tree.scrollTop = sys.getStorage('scrollTop') || 0;
    });
}

function resetHeight() {
    setTimeout(function () {
        var maxHeight = 0,
            height = 0;
        if (allShowItemsCount > maxItemsCount) {
            maxHeight = maxItemsCount * unitItemHeight;
        } else {
            maxHeight = allShowItemsCount * unitItemHeight + maxItemsMargin;
        }
        height = Math.max(minWindowHeight, maxHeight);
        sys.setStorage('popupHeight', height);
        document.body.style.height = ui.$('resize').style.height = height + 'px';
    }, 100);
};

var toolbar = {
    resetWidth: function (width) {
        ui.$('toolbar').style.width = width + 'px';
        ui.$('searchInput').style.width = width - 114 + 'px';
    },
    resetTop: function (height) {
        ui.$('toolbar').style.top = height + 'px';
    }
}

var menu = {
    data: {
        folder: [{
                name: 'openAll',
                act: 'openFolderInCurWindow'
            },
            {
                name: 'openAllInNewWindow',
                act: 'openFolderInNewWindow'
            },
            {
                name: 'openAllInIncognitoWindow',
                act: 'openFolderInIncognitoWindow'
            },
            /*{name:'line', act:''},
            {name:'rename', act:'rename'},*/
            {
                name: 'line',
                act: ''
            },
            {
                name: 'delete',
                act: 'removeTree'
            },
            {
                name: 'line',
                act: ''
            },
            {
                name: 'manager',
                act: 'manager'
            }
        ],
        file: [{
                name: 'openItem',
                act: 'openItem'
            },
            {
                name: 'openInIncognitoWindow',
                act: 'openInIncognitoWindow'
            },
            {
                name: 'line',
                act: ''
            },
            /*{name:'modify', act:'modify'},
            {name:'replace', act:'replace'},*/
            {
                name: 'line',
                act: ''
            },
            {
                name: 'delete',
                act: 'remove'
            },
            {
                name: 'line',
                act: ''
            },
            {
                name: 'manager',
                act: 'manager'
            }
        ]
    },
    tip: null,
    item: null,
    folderMenu: null,
    fileMenu: null,
    size: {
        width: 160,
        height: 0
    },
    init: function () {
        menu.create('file');
        menu.create('folder');
        document.oncontextmenu = function (e) {
            var ele = e.target;
            // 修复在右键菜单项中再次点击鼠标右键时会出错的问题
            // 2020-11-15
            // pana
            if ((ele.tagName == 'LI') && (!ele.closest('div') || !ele.closest('div').classList.contains('menu'))) {
                menu.hideMenu();
                menu.item = {
                    id: ele.id,
                    parentId: ele.parentId,
                    url: ele.url,
                    type: ele.type,
                    status: ele.status
                };
                menu.showMenu(e);
                e.preventDefault();
            }
        };
    },
    create: function (type) {
        var menuUl = ui.create('ul'),
            menuData = menu.data[type],
            isLine = false,
            className = '',
            itemCount = 0;
        menu.tip = ui.create("div");
        menu.tip = ui.create("div");
        menu.tip.id = "menu-" + type;
        menu.tip.className = "menu";
        menu.tip.onmousedown = function (e) {
            e.stopPropagation();
        };

        for (i in menuData) {
            if (isLine) {
                className = 'menu-line';
                isLine = false;
            }
            if (menuData[i].name == 'line') {
                isLine = true;
            } else {
                var menuLi = ui.create('li');
                menuLi.className = className;
                menuLi.id = 'menu_' + menuData[i].name;
                menuLi.textContent = chrome.i18n.getMessage(menuData[i].name);
                let funcName = menuData[i].act;
                menuLi.addEventListener('click', function(e) {
                    menu.exec(funcName);
                });
                menuUl.appendChild(menuLi);
                className = '';
                itemCount++;
            }
        }

        menu.size.height = 21 * itemCount;
        menu.tip.style.height = menu.size.height + 4 + 'px';
        menu.tip.style.width = menu.size.width + 'px';
        menu.tip.appendChild(menuUl);
        menu[type + 'Menu'] = menu.tip;
        document.body.appendChild(menu.tip);
    },
    exec: function (funcName) {
        menuOp[menu.item.type][funcName]();
        menu.hideMenu();
        menu.item = null;
    },
    showMenu: function (e) {
        e.target.style.background = '#ccc';
        e.target.style.webkitBorderRadius = '3px';
        var mousePos = ui.getMousePos(e),
            screenSize = {
                'width': document.documentElement.clientWidth - 5,
                'height': document.documentElement.clientHeight - 5
            },
            left = document.body.scrollLeft + mousePos.x,
            top = document.body.scrollTop + mousePos.y;

        left = (screenSize.width - mousePos.x < menu.size.width) ? left - menu.size.width : left;
        top = (screenSize.height - mousePos.y < menu.size.height) ? top - menu.size.height : top;
        menu.tip = menu[e.target.type + 'Menu'];
        menu.tip.style.left = left + "px";
        menu.tip.style.top = top + "px";
        menu.tip.style.display = "block";
    },
    hideMenu: function () {
        if (menu.item && ui.$(menu.item.id)) {
            ui.$(menu.item.id).style.background = 'none';
        }
        menu.tip.style.display = "none";
        //��ֹmenu�ĸ��ţ������Ƴ���Ч������
        menu.tip.style.left = "-50px";
        menu.tip.style.top = "-50px";
    }
};

var menuOp = {
    isGotCurFiles: false,
    getCurFiles: function (folderID, callback) {
        var files = [],
            children = sys.getChildren(folderID);
        children.forEach(function (child) {
            if (!Array.isArray(child.children)) {
                files.push(child);
            }
        });
        callback(files);
    },
    folder: {
        openFolderInCurWindow: function () {
            var folderID = ui.getItemID(menu.item);
            menuOp.getCurFiles(folderID, function (items) {
                items.forEach(function (item) {
                    actions.openItem(item, false);
                });
            });
        },
        openItemsInWindow: function (isIncognito) {
            var folderID = ui.getItemID(menu.item);
            menuOp.getCurFiles(folderID, function (items) {
                actions.openItemsInWindow(items, isIncognito);
            });
        },
        openFolderInNewWindow: function () {
            menuOp.folder.openItemsInWindow(false);
        },
        openFolderInIncognitoWindow: function () {
            menuOp.folder.openItemsInWindow(true);
        },
        rename: function () {

        },
        removeTree: function () {
            actions.removeTree(menu.item);
        },
        manager: function () {
            actions.manager();
        }
    },
    file: {
        openItem: function () {
            actions.openItem(menu.item, false);
        },
        openInIncognitoWindow: function () {
            actions.openItemsInWindow([menu.item], true);
        },
        modify: function () {

        },
        replace: function () {

        },
        remove: function () {
            actions.removeItem(menu.item);
        },
        manager: function () {
            actions.manager();
        }
    }
};

initTree();
menu.init();

resizer.addEventListener('mousedown', function (e) {
    e.preventDefault();
    e.stopPropagation();
    resizeHandle = true;
    bodyWidth = document.body.offsetWidth;
    screenX = e.screenX;
});
tree.addEventListener('click', function () {
    resetHeight();
    menu.hideMenu();
});
tree.addEventListener('scroll', function () {
    //������ģʽ ��¼������λ��
    !searchValue && sys.setStorage('scrollTop', tree.scrollTop);
    menu.hideMenu();
});
document.body.addEventListener('mousedown', function () {
    menu.hideMenu();
});
document.addEventListener('mousemove', function (e) {
    if (resizeHandle) {
        e.preventDefault();
        var width = bodyWidth + e.screenX - screenX,
            maxWidth = Math.min(640, Math.max(320, width));
        document.body.style.width = maxWidth + 'px';
        resizer.style.left = maxWidth - 5 + 'px';
        toolbar.resetWidth(maxWidth);
    }
});
document.addEventListener('mouseup', function (e) {
    e.preventDefault();
    resizeHandle = false;
    var bodyWidth = document.body.style.width.slice(0, -2);
    sys.setStorage('bodyWidth', bodyWidth);
    toolbar.resetWidth(bodyWidth);
});

// Disable Chrome auto-scroll feature
window.addEventListener('mousedown', function (e) {
    if (e.button == 1) e.preventDefault();
});
searchInput.placeholder = chrome.i18n.getMessage('searchDefaultValue');
searchInput.addEventListener('keyup', function (e) {
    var value = this.value.trim();
    if (searchValue != value) {
        searchValue = value;
        buildSearchResult(searchValue);
    }
});
searchInput.addEventListener('click', function (e) {
    if (this.value.trim() == '' && searchValue != '') {
        initTree();
    }
});

ui.$('managerButton').addEventListener('click', actions.manager);